//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) - 2013 - Scilab Enterprises - Clement David
// Copyright (C) - 2013 - Scilab Enterprises - Vladislav Trubkin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
//
#include <stdlib.h>
#include <stdio.h>

#include "api_scilab.h"
#include "Scierror.h"
#include "localization.h"
#include "sciprint.h"

#include "fmu_wrapper.h"
#include "fmiModelFunctions.h"

/*
 * Declare all the fmu_call_wrapper
 */
#define FMI_KLASS_METHOD(fun) static int fmu_call_ ## fun (const FmuModelExchange* instance);
FMI_KLASS_METHOD(fmiGetModelTypesPlatform)
FMI_KLASS_METHOD(fmiGetVersion)
FMI_KLASS_METHOD(fmiInstantiateModel)
FMI_KLASS_METHOD(fmiFreeModelInstance)
FMI_KLASS_METHOD(fmiSetDebugLogging)
FMI_KLASS_METHOD(fmiSetTime)
FMI_KLASS_METHOD(fmiSetContinuousStates)
FMI_KLASS_METHOD(fmiCompletedIntegratorStep)
FMI_KLASS_METHOD(fmiSetReal)
FMI_KLASS_METHOD(fmiSetInteger)
FMI_KLASS_METHOD(fmiSetBoolean)
FMI_KLASS_METHOD(fmiSetString)
FMI_KLASS_METHOD(fmiInitialize)
FMI_KLASS_METHOD(fmiGetDerivatives)
FMI_KLASS_METHOD(fmiGetEventIndicators)
FMI_KLASS_METHOD(fmiGetReal)
FMI_KLASS_METHOD(fmiGetInteger)
FMI_KLASS_METHOD(fmiGetBoolean)
FMI_KLASS_METHOD(fmiGetString)
FMI_KLASS_METHOD(fmiEventUpdate)
FMI_KLASS_METHOD(fmiGetContinuousStates)
FMI_KLASS_METHOD(fmiGetNominalContinuousStates)
FMI_KLASS_METHOD(fmiGetStateValueReferences)
FMI_KLASS_METHOD(fmiTerminate)
#undef FMI_KLASS_METHOD

/*
 * Table to map each fmu_call_XXXX name to gateway
 */
typedef int fmu_call_TYPE(const FmuModelExchange* instance);
struct fmu_call_map
{
    const char* fname;
    fmu_call_TYPE* funptr;
};

static const struct fmu_call_map nameFunMap[] =
{
#define FMI_KLASS_METHOD(fun) { #fun , fmu_call_ ## fun },
    FMI_KLASS_METHOD(fmiGetModelTypesPlatform)
    FMI_KLASS_METHOD(fmiGetVersion)
    FMI_KLASS_METHOD(fmiInstantiateModel)
    FMI_KLASS_METHOD(fmiFreeModelInstance)
    FMI_KLASS_METHOD(fmiSetDebugLogging)
    FMI_KLASS_METHOD(fmiSetTime)
    FMI_KLASS_METHOD(fmiSetContinuousStates)
    FMI_KLASS_METHOD(fmiCompletedIntegratorStep)
    FMI_KLASS_METHOD(fmiSetReal)
    FMI_KLASS_METHOD(fmiSetInteger)
    FMI_KLASS_METHOD(fmiSetBoolean)
    FMI_KLASS_METHOD(fmiSetString)
    FMI_KLASS_METHOD(fmiInitialize)
    FMI_KLASS_METHOD(fmiGetDerivatives)
    FMI_KLASS_METHOD(fmiGetEventIndicators)
    FMI_KLASS_METHOD(fmiGetReal)
    FMI_KLASS_METHOD(fmiGetInteger)
    FMI_KLASS_METHOD(fmiGetBoolean)
    FMI_KLASS_METHOD(fmiGetString)
    FMI_KLASS_METHOD(fmiEventUpdate)
    FMI_KLASS_METHOD(fmiGetContinuousStates)
    FMI_KLASS_METHOD(fmiGetNominalContinuousStates)
    FMI_KLASS_METHOD(fmiGetStateValueReferences)
    FMI_KLASS_METHOD(fmiTerminate)
#undef FMI_KLASS_METHOD
    {
        NULL, NULL
    }
};

int fmu_call(char *fname, unsigned long fname_len)
{
    SciErr sciErr;
    int i;

    int *piAddressFmuInstance;
    int *piAddressFunctionName;
    void* pInstance;
    FmuModelExchange* instance;
    char *varFunctionName;

    fmu_call_TYPE* funptr;
    int nbOutputVariable;

    /*
     * The first 2 input arguments are : instance (pointer), function name string
     */
    if (!checkInputArgumentAtLeast(pvApiCtx, 2))
    {
        return 0;
    }
    sciErr = getVarAddressFromPosition(pvApiCtx, 1, &piAddressFmuInstance);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        Scierror(999, _("%s: Can not read input argument #%d.\n"), fname, 1);
        return -sciErr.iErr;
    }
    sciErr = getVarAddressFromPosition(pvApiCtx, 2, &piAddressFunctionName);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        Scierror(999, _("%s: Can not read input argument #%d.\n"), fname, 2);
        return -sciErr.iErr;
    }
    sciErr = getPointer(pvApiCtx, piAddressFmuInstance, &pInstance) ;
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        Scierror(999, _("%s: Can not read input argument #%d.\n"), fname, 1);
        return -sciErr.iErr;
    }
    instance = fmu_wrap_model_exchange(pInstance);
    if (instance == NULL)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU expected.\n"), fname, 1);
        return -999;
    }
    sciErr.iErr = getAllocatedSingleString(pvApiCtx, piAddressFunctionName, &varFunctionName);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Can not read input argument #%d.\n"), fname, 2);
        return -999;
    }

    /*
     * Locate the local gateway
     */
    for (i = 0, funptr = NULL; nameFunMap[i].fname != NULL; i++)
    {
        if (strcmp(nameFunMap[i].fname, varFunctionName) == 0)
        {
            funptr = nameFunMap[i].funptr;
            break;
        }
    }
    if (funptr == NULL)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU function name expected.\n"), fname, 2);
        free(varFunctionName);
        return -999;
    }

    nbOutputVariable = funptr(instance);
    if (nbOutputVariable < 0)
    {
        Scierror(999, _("%s: Error while calling %s.\n"), fname, varFunctionName);
        return -999;
    }

    /*
     * Assign all the output variables
     */
    for (i = 1; i <= nbOutputVariable; i++)
    {
        AssignOutputVariable(pvApiCtx, i) = nbInputArgument(pvApiCtx) + i;
    }
    return 0;
}

///**
// * log some information into the Scilab console
// */
//static void fmu_callback_logger(fmiComponent c, fmiString instanceName, fmiStatus status, fmiString category, fmiString message, fmiBoolean loggingOn, ...)
//void fmu_callback_logger(fmiComponent c, fmiString instanceName, fmiStatus status, fmiString category, fmiString message, fmiBoolean loggingOn, ...)
//{
//    va_list ap;
//
//    if (!instanceName)
//    {
//        instanceName = "?";
//    }
//    if (!category)
//    {
//        category = "?";
//    }
//    if ((!status) && strcmp(message, "fmiSetDebugLogging") && (!loggingOn)) 
//    {
//        return;
//    }
//    
//    sciprint("fmiStatus = %d\n", status);
//    sciprint("%s (%s)\n", instanceName, category);
//    va_start(ap, message);
//    scivprint(message, ap);
//    va_end(ap);
//
//    /* FMI for ModelExchange v1.0: "logger function will append a line break to each message" */
//    sciprint("\n");
//}

struct CommonAccessorsModifiers
{
    const char* funName;
    BOOL calculateNvr;
    int vrRows;
    int vrCols;
};

/**
 * Utility to set/get some values by valueReferences
 * @param[in] instance the model class (FMU)
 * @param[out] c the component (FMU instance)
 * @param[out] vr the value references to get
 * @param[out] nvr length of vr
 * @param[in/out] modifiers meta-data
 * @return if the operation succeeded (0) or not ( !0 )
 */
static int fmiCommonAccessors(const FmuModelExchange* instance, fmiComponent* c, fmiValueReference** vr, size_t* nvr, struct CommonAccessorsModifiers* modifier);

/**
 * Utility to create an eventInfo mlist into the stack
 * @param[in] _pvCtx stack context opaque pointer
 * @param[in] _iVar variable number
 * @param[in] eventInfo value
 * @return if the operation succeeded (0) or not ( !0 )
 */
static int createEventInfoMList(void* _pvCtx, int _iVar, fmiEventInfo eventInfo);

/*
 * Implementation of the fmu_call_XXX methods
 */
#define FMI_KLASS_METHOD(fun) static int fmu_call_ ## fun (const FmuModelExchange* instance)

FMI_KLASS_METHOD(fmiGetModelTypesPlatform)
{
    SciErr sciErr;
    const char* typePlatform;

    if (!checkInputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    typePlatform = instance->_fmiGetModelTypesPlatform();

    sciErr.iErr = createSingleString(pvApiCtx, nbInputArgument(pvApiCtx) + 1, typePlatform);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiGetVersion)
{
    SciErr sciErr;
    const char* version;

    if (!checkInputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    version = instance->_fmiGetVersion();

    sciErr.iErr = createSingleString(pvApiCtx, nbInputArgument(pvApiCtx) + 1, version);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiInstantiateModel)
{
    SciErr sciErr;

    int *piAddressInstanceName;
    int *piAddressGUID;
    int *piAddressLoggingOn;

    fmiCallbackFunctions functions;
    int scilabLoggingOn;

    /* should be fmiString but avoid const usage there */
    char* instanceName;
    /* should be fmiString but avoid const usage there */
    char* GUID;
    fmiBoolean loggingOn;

    fmiComponent component;

    if (!checkInputArgument(pvApiCtx, 4, 5))
    {
        return -1;
    }

    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressInstanceName);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getAllocatedSingleString(pvApiCtx, piAddressInstanceName, &instanceName);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAddressGUID);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        freeAllocatedSingleString(instanceName);
        return -sciErr.iErr;
    }

    sciErr.iErr = getAllocatedSingleString(pvApiCtx, piAddressGUID, &GUID);
    if (sciErr.iErr)
    {
        freeAllocatedSingleString(instanceName);
        return -sciErr.iErr;
    }

    /* loggingOn is an optional argument */
    if (nbInputArgument(pvApiCtx) > 4)
    {
        sciErr = getVarAddressFromPosition(pvApiCtx, 5, &piAddressLoggingOn);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            freeAllocatedSingleString(instanceName);
            freeAllocatedSingleString(GUID);
            return -sciErr.iErr;
        }

        sciErr.iErr = getScalarBoolean(pvApiCtx, piAddressLoggingOn, &scilabLoggingOn);
        if (sciErr.iErr)
        {
            freeAllocatedSingleString(instanceName);
            freeAllocatedSingleString(GUID);
            return -sciErr.iErr;
        }

        loggingOn = scilabLoggingOn == TRUE ? fmiTrue : fmiFalse;
    }
    else
    {
        loggingOn = fmiTrue;
    }

    functions.logger = fmu_callback_logger;
    functions.allocateMemory = calloc;
    functions.freeMemory = free;
    functions.stepFinished = NULL;
    
    component = instance->_fmiInstantiateModel(instanceName, GUID, functions, loggingOn);

    if (!component)
    {
        fmu_wrap_model_exchange_delete(instance);
        Scierror(999, _("%s: %s model instantiation failed.\n"), "fmiInstantiateModel", instanceName);
        freeAllocatedSingleString(instanceName);
        freeAllocatedSingleString(GUID);
        return 0;
    }
    
    freeAllocatedSingleString(instanceName);
    freeAllocatedSingleString(GUID);
    
    sciErr = createPointer(pvApiCtx, nbInputArgument(pvApiCtx) + 1, component);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiFreeModelInstance)
{
    SciErr sciErr;

    int *piAddressC;

    fmiComponent c;

    if (!checkInputArgument(pvApiCtx, 3, 3))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 0, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiFreeModelInstance", 3);
        return 0;
    }

    instance->_fmiFreeModelInstance(c);
    return 0;
}

FMI_KLASS_METHOD(fmiSetDebugLogging)
{
    SciErr sciErr;

    int *piAddressC;
    int *piAddressLoggingOn;

    int scilabLoggingOn;

    fmiComponent c;
    fmiBoolean loggingOn;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiSetDebugLogging", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAddressLoggingOn);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getScalarBoolean(pvApiCtx, piAddressLoggingOn, &scilabLoggingOn);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    loggingOn = scilabLoggingOn == TRUE ? fmiTrue : fmiFalse;

    status = instance->_fmiSetDebugLogging(c, loggingOn);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiSetTime)
{
    SciErr sciErr;

    int *piAddressC;
    int *piAdressTime;

    fmiComponent c;
    fmiReal time;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiSetTime", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressTime);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getScalarDouble(pvApiCtx, piAdressTime, &time);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    status = instance->_fmiSetTime(c, time);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiSetContinuousStates)
{
    SciErr sciErr;

    int *piAddressC;
    int *piAdressContinuousStates;
    int rows;
    int cols;

    fmiComponent c;
    fmiReal* x;
    size_t nx;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 5, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiSetContinuousStates", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressContinuousStates);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( !isDoubleType(pvApiCtx, piAdressContinuousStates) )
    {
        Scierror(999, _("%s: Wrong type for argument %d: Matrix expected.\n"), "fmiSetContinuousStates", 4);
        return 0;
    }

    sciErr = getMatrixOfDouble(pvApiCtx, piAdressContinuousStates, &rows, &cols, &x);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    nx = rows * cols;
    status = instance->_fmiSetContinuousStates(c, x, nx);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiCompletedIntegratorStep)
{
    SciErr sciErr;

    int *piAddressComponent;
    int scilabCallEventUpdate;

    fmiComponent c;

    fmiStatus status;
    fmiBoolean callEventUpdate;

    if (!checkInputArgument(pvApiCtx, 3, 3))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiCompletedIntegratorStep", 3);
        return 0;
    }

    status = instance->_fmiCompletedIntegratorStep(c, &callEventUpdate);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    scilabCallEventUpdate = callEventUpdate == fmiTrue ? TRUE : FALSE;
    sciErr.iErr = createScalarBoolean(pvApiCtx, nbInputArgument(pvApiCtx) + 2, scilabCallEventUpdate);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 2;
}

FMI_KLASS_METHOD(fmiSetReal)
{
    SciErr sciErr;

    int* piAdressValue;
    int rows;
    int cols;

    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;
    fmiReal* value;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 5, 6))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    modifiers.funName = "fmiSetReal";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 5;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, nbInputArgument(pvApiCtx), &piAdressValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = ! isDoubleType(pvApiCtx, piAdressValue);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for argument %d: double matrix expected.\n"), "fmiSetReal", nbInputArgument(pvApiCtx));
        return 0;
    }

    sciErr = getMatrixOfDouble(pvApiCtx, piAdressValue, &rows, &cols, &value);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( nvr != rows * cols )
    {
        Scierror(999, _("%s: Incompatible input arguments #%d and #%d: Same sizes expected.\n"), "fmiSetReal", 4, nbInputArgument(pvApiCtx));
        return 0;
    }

    status = instance->_fmiSetReal(c, vr, nvr, value);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiSetInteger)
{
    SciErr sciErr;

    int* piAdressValue;
    int rows;
    int cols;

    BOOL valueHasIntegerType;
    int valueIntegerType;
    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;
    fmiInteger* value;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 5, 6))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    modifiers.funName = "fmiSetInteger";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 5;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, nbInputArgument(pvApiCtx), &piAdressValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    valueHasIntegerType = isIntegerType(pvApiCtx, piAdressValue);
    if (valueHasIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressValue, &valueIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        valueIntegerType = -1;
    }

    if (valueHasIntegerType)
    {
        switch (valueIntegerType)
        {
            case SCI_INT32:
                sciErr = getMatrixOfInteger32(pvApiCtx, piAdressValue, &rows, &cols, &value);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiSetInteger", nbInputArgument(pvApiCtx));
                return -1;
                break;
        }
    }
    else
    {
        sciErr = getMatrixOfDoubleAsInteger(pvApiCtx, piAdressValue, &rows, &cols, &value);
    }
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( nvr != rows * cols )
    {
        Scierror(999, _("%s: Incompatible input arguments #%d and #%d: Same sizes expected.\n"), "fmiSetInteger", 4, nbInputArgument(pvApiCtx));
        return 0;
    }

    status = instance->_fmiSetInteger(c, vr, nvr, value);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiSetBoolean)
{
    SciErr sciErr;

    int* piAdressValue;
    int rows = 0;
    int cols = 0;

    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;
    fmiBoolean* value;

    fmiStatus status;

    int* scilabValue;
    int i;

    if (!checkInputArgument(pvApiCtx, 5, 6))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    modifiers.funName = "fmiSetBoolean";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 5;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, nbInputArgument(pvApiCtx), &piAdressValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = ! isBooleanType(pvApiCtx, piAdressValue);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for argument %d: boolean matrix expected.\n"), "fmiSetBoolean", nbInputArgument(pvApiCtx));
        return 0;
    }

    sciErr = getMatrixOfBoolean(pvApiCtx, piAdressValue, &rows, &cols, &scilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( nvr != rows * cols )
    {
        Scierror(999, _("%s: Incompatible input arguments #%d and #%d: Same sizes expected.\n"), "fmiSetBoolean", 4, nbInputArgument(pvApiCtx));
        return 0;
    }

    /* Copy the value has in Scilab a boolean is stored as an int whereas in the FMI a boolean is a char */
    value = malloc(nvr * sizeof(fmiBoolean));
    for (i = 0; i < nvr; i++)
    {
        value[i] = scilabValue[i];
    }

    status = instance->_fmiSetBoolean(c, vr, nvr, value);

    free(value);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiSetString)
{
    SciErr sciErr;

    int* piAdressValue;
    int rows = 0;
    int cols = 0;

    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;
    /* fmiString not used on purpose: avoid (const char*) to (char*) cast */
    char** value;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 5, 6))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    modifiers.funName = "fmiSetString";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 5;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, nbInputArgument(pvApiCtx), &piAdressValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( ! isStringType(pvApiCtx, piAdressValue) )
    {

        Scierror(999, _("%s: Wrong type for argument %d: string matrix expected.\n"), "fmiSetString", nbInputArgument(pvApiCtx));
        return 0;
    }

    sciErr.iErr = getAllocatedMatrixOfString(pvApiCtx, piAdressValue, &rows, &cols, &value);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    if ( nvr != rows * cols )
    {
        Scierror(999, _("%s: Incompatible input arguments #%d and #%d: Same sizes expected.\n"), "fmiSetString", 4, nbInputArgument(pvApiCtx));
        freeAllocatedMatrixOfString(rows, cols, value);
        return 0;
    }

    /* from the spec: All strings passed as arguments to fmiSetString must be copied inside this function */
    status = instance->_fmiSetString(c, vr, nvr, (fmiString*) value);

    freeAllocatedMatrixOfString(rows, cols, value);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

FMI_KLASS_METHOD(fmiInitialize)
{
    SciErr sciErr;

    int* piAddressC;
    int* piAdressToleranceControlled;
    int* piAdressRelativeTolerance;

    fmiComponent c;
    int scilabToleranceControlled;
    fmiBoolean toleranceControlled;
    fmiReal relativeTolerance;

    fmiEventInfo eventInfo;
    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 5, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiInitialize", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressToleranceControlled);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getScalarBoolean(pvApiCtx, piAdressToleranceControlled, &scilabToleranceControlled);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    if (scilabToleranceControlled == TRUE)
    {
        toleranceControlled = fmiTrue;
    }
    else
    {
        toleranceControlled = fmiFalse;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 5, &piAdressRelativeTolerance);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getScalarDouble(pvApiCtx, piAdressRelativeTolerance, &relativeTolerance);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    status = instance->_fmiInitialize(c, toleranceControlled, relativeTolerance, &eventInfo);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr.iErr = createEventInfoMList(pvApiCtx, nbInputArgument(pvApiCtx) + 2, eventInfo);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 2;
}

FMI_KLASS_METHOD(fmiGetDerivatives)
{
    SciErr sciErr;

    int* piAddressVarComponent;
    int* piAdressVarNx;

    BOOL nxHasAValidIntegerType;
    int nxIntegerType;
    double dblNx;
    double* pdblStatus;

    fmiComponent c;
    /* should be a size_t, but avoid some casts */
    int nx;

    fmiStatus status;
    fmiReal* derivatives;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiGetDerivatives", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVarNx);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    nxHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVarNx);
    if (nxHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVarNx, &nxIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        nxIntegerType = -1;
    }

    if (nxHasAValidIntegerType)
    {
        switch (nxIntegerType)
        {
            case SCI_INT32:
                sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressVarNx, &nx);
                break;
            case SCI_UINT32:
                sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressVarNx, (unsigned int*) &nx);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiGetDerivatives", 4);
                return 0;
                break;
        }
    }
    else
    {
        sciErr.iErr = getScalarDouble(pvApiCtx, piAdressVarNx, &dblNx);
        nx = (int) dblNx;
    }
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 2, nx, 1, &derivatives);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetDerivatives(c, derivatives, nx);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetEventIndicators)
{
    SciErr sciErr;

    int* piAddressVarComponent;
    int* piAdressVarNi;

    BOOL niHasAValidIntegerType;
    int niIntegerType;
    double dblNi;
    double* pdblStatus;

    fmiComponent c;
    /* should be a size_t, but avoid some casts */
    int ni;

    fmiStatus status;
    fmiReal* eventIndicators;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiGetEventIndicators", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVarNi);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    niHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVarNi);
    if (niHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVarNi, &niIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        niIntegerType = -1;
    }

    if (niHasAValidIntegerType)
    {
        switch (niIntegerType)
        {
            case SCI_INT32:
                sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressVarNi, &ni);
                break;
            case SCI_UINT32:
                sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressVarNi, (unsigned int*) &ni);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiGetEventIndicators", 4);
                return 0;
                break;
        }
    }
    else
    {
        sciErr.iErr = getScalarDouble(pvApiCtx, piAdressVarNi, &dblNi);
        ni = (int) dblNi;

    }
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 2, ni, 1, &eventIndicators);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetEventIndicators(c, eventIndicators, ni);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetReal)
{
    SciErr sciErr;

    double* pdblStatus;
    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;

    fmiStatus status;
    fmiReal* value;

    if (!checkInputArgument(pvApiCtx, 4, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    modifiers.funName = "fmiGetReal";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 4;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return 0;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 2, modifiers.vrRows, modifiers.vrCols, &value);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetReal(c, vr, nvr, value);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetInteger)
{
    SciErr sciErr;

    double* pdblStatus;
    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;

    fmiStatus status;
    fmiInteger* value;

    if (!checkInputArgument(pvApiCtx, 4, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    modifiers.funName = "fmiGetInteger";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 4;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfInteger32(pvApiCtx, nbInputArgument(pvApiCtx) + 2, modifiers.vrRows, modifiers.vrCols, &value);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetInteger(c, vr, nvr, value);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetBoolean)
{
    SciErr sciErr;

    int i;
    int* scilabValue;
    double* pdblStatus;
    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;

    fmiStatus status;
    fmiBoolean* value;


    if (!checkInputArgument(pvApiCtx, 4, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    modifiers.funName = "fmiGetBoolean";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 4;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    value = malloc(nvr * sizeof(fmiBoolean));
    if (!value)
    {
        Scierror(999, _("%s: Memory allocation error.\n"), "fmiGetBoolean");
        return 0;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        free(value);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfBoolean(pvApiCtx, nbInputArgument(pvApiCtx) + 2, modifiers.vrRows, modifiers.vrCols, &scilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        free(value);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetBoolean(c, vr, nvr, value);

    *pdblStatus = status;

    for (i = 0; i < nvr; i++)
    {
        scilabValue[i] = value[i];
    }
    free(value);

    return 2;
}

FMI_KLASS_METHOD(fmiGetString)
{
    SciErr sciErr;

    struct CommonAccessorsModifiers modifiers;

    fmiComponent c;
    fmiValueReference* vr;
    size_t nvr;

    fmiStatus status;
    fmiString* value;

    if (!checkInputArgument(pvApiCtx, 4, 5))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    modifiers.funName = "fmiGetString";
    modifiers.calculateNvr = nbInputArgument(pvApiCtx) == 4;
    sciErr.iErr = fmiCommonAccessors(instance, &c, &vr, &nvr, &modifiers);
    if (sciErr.iErr)
    {
        return 0;
    }

    value = malloc(nvr * sizeof(fmiString));
    if (!value)
    {
        Scierror(999, _("%s: Memory allocation error.\n"), "fmiGetString");
        return 0;
    }

    status = instance->_fmiGetString(c, vr, nvr, value);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        free(value);
        return -sciErr.iErr;
    }

    sciErr = createMatrixOfString(pvApiCtx, nbInputArgument(pvApiCtx) + 2, modifiers.vrRows, modifiers.vrCols, value);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        free(value);
        return -sciErr.iErr;
    }

    free(value);
    return 2;
}

FMI_KLASS_METHOD(fmiEventUpdate)
{
    SciErr sciErr;

    int* piAddressC;
    int* piAdressIntermediateResults;

    fmiComponent c;
    int scilabIntermediateResults;
    fmiBoolean intermediateResults;

    fmiEventInfo eventInfo;
    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiEventUpdate", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressIntermediateResults);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr.iErr = getScalarBoolean(pvApiCtx, piAdressIntermediateResults, &scilabIntermediateResults);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    intermediateResults = scilabIntermediateResults == TRUE ? fmiTrue : fmiFalse;
    status = instance->_fmiEventUpdate(c, intermediateResults, &eventInfo);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr.iErr = createEventInfoMList(pvApiCtx, nbInputArgument(pvApiCtx) + 2, eventInfo);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 2;
}

FMI_KLASS_METHOD(fmiGetContinuousStates)
{
    SciErr sciErr;

    int* piAddressVarComponent;
    int* piAdressVarNx;

    BOOL nxHasAValidIntegerType;
    int nxIntegerType;
    double dblNx;
    double* pdblStatus;

    fmiComponent c;
    /* should be a size_t, but avoid some casts */
    int nx;

    fmiStatus status;
    fmiReal* x;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiGetContinuousStates", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVarNx);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    nxHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVarNx);
    if (nxHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVarNx, &nxIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        nxIntegerType = -1;
    }

    if (nxHasAValidIntegerType)
    {
        switch (nxIntegerType)
        {
            case SCI_INT32:
                sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressVarNx, &nx);
                break;
            case SCI_UINT32:
                sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressVarNx, (unsigned int*) &nx);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiGetContinuousStates", 4);
                return 0;
                break;
        }
    }
    else
    {
        sciErr.iErr = getScalarDouble(pvApiCtx, piAdressVarNx, &dblNx);
        nx = (int) dblNx;
    }
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 2, nx, 1, &x);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetContinuousStates(c, x, nx);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetNominalContinuousStates)
{
    SciErr sciErr;

    int* piAddressVarComponent;
    int* piAdressVarNx;

    BOOL nxHasAValidIntegerType;
    int nxIntegerType;
    double dblNx;
    double* pdblStatus;

    fmiComponent c;
    /* should be a size_t, but avoid some casts */
    int nx;

    fmiStatus status;
    fmiReal* x_nominal;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiGetNominalContinuousStates", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVarNx);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    nxHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVarNx);
    if (nxHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVarNx, &nxIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        nxIntegerType = -1;
    }

    if (nxHasAValidIntegerType)
    {
        switch (nxIntegerType)
        {
            case SCI_INT32:
                sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressVarNx, &nx);
                break;
            case SCI_UINT32:
                sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressVarNx, (unsigned int*) &nx);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiGetContinuousStates", 4);
                return 0;
                break;
        }
    }
    else
    {
        sciErr.iErr = getScalarDouble(pvApiCtx, piAdressVarNx, &dblNx);
        nx = (int) dblNx;
    }
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 2, nx, 1, &x_nominal);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetNominalContinuousStates(c, x_nominal, nx);
    *pdblStatus = status;
    return 2;
}

FMI_KLASS_METHOD(fmiGetStateValueReferences)
{
    SciErr sciErr;

    int* piAddressVarComponent;
    int* piAdressVarNx;

    BOOL nxHasAValidIntegerType;
    int nxIntegerType;
    double dblNx;
    double* pdblStatus;

    fmiComponent c;
    /* should be a size_t, but avoid some casts */
    int nx;

    fmiStatus status;
    /* should be a fmiValueReference, but avoid some casts */
    int* vrx;

    if (!checkInputArgument(pvApiCtx, 4, 4))
    {
        return -1;
    }
    if (!checkOutputArgument(pvApiCtx, 2, 2))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiGetStateValueReferences", 3);
        return 0;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVarNx);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    nxHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVarNx);
    if (nxHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVarNx, &nxIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return -sciErr.iErr;
        }
    }
    else
    {
        nxIntegerType = -1;
    }

    if (nxHasAValidIntegerType)
    {
        switch (nxIntegerType)
        {
            case SCI_INT32:
                sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressVarNx, &nx);
                break;
            case SCI_UINT32:
                sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressVarNx, (unsigned int*) &nx);
                break;
            default:
                Scierror(999, _("%s: Wrong value for input argument #%d: An integer value expected.\n"), "fmiGetStateValueReferences", 4);
                return 0;
                break;
        }
    }
    else
    {
        sciErr.iErr = getScalarDouble(pvApiCtx, piAdressVarNx, &dblNx);
        nx = (int) dblNx;
    }
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, 1, 1, &pdblStatus);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = allocMatrixOfDoubleAsInteger(pvApiCtx, nbInputArgument(pvApiCtx) + 2, nx, 1, &vrx);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    status = instance->_fmiGetStateValueReferences(c, (fmiValueReference*) vrx, nx);
    *pdblStatus = status;
    return 2;
}


FMI_KLASS_METHOD(fmiTerminate)
{
    SciErr sciErr;

    int* piAddressVarComponent;

    fmiComponent c;

    fmiStatus status;

    if (!checkInputArgument(pvApiCtx, 3, 3))
    {
        return -1;

    }
    if (!checkOutputArgument(pvApiCtx, 1, 1))
    {
        return -1;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressVarComponent);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return -sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressVarComponent, &c);
    if (sciErr.iErr)
    {
        Scierror(999, _("%s: Wrong type for input argument #%d: FMU Component expected.\n"), "fmiTerminate", 3);
        return 0;
    }

    status = instance->_fmiTerminate(c);

    sciErr.iErr = createScalarDouble(pvApiCtx, nbInputArgument(pvApiCtx) + 1, status);
    if (sciErr.iErr)
    {
        return -sciErr.iErr;
    }
    return 1;
}

#undef FMI_KLASS_METHOD

static int fmiCommonAccessors(const FmuModelExchange* instance, fmiComponent* c, fmiValueReference** vr, size_t* nvr, struct CommonAccessorsModifiers* modifier)
{
    SciErr sciErr;

    int* piAddressC;
    int* piAdressVr;
    int* piAdressNvr;
    int rows;
    int cols;

    BOOL vrHasAValidIntegerType;
    int vrIntegerType;
    BOOL nvrHasIntegerType;
    int nvrIntegerType;

    sciErr = getVarAddressFromPosition(pvApiCtx, 3, &piAddressC);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }

    sciErr = getPointer(pvApiCtx, piAddressC, c);
    if (sciErr.iErr)
    {
        sciprint(_("%s: Wrong type for input argument #%d: FMU Component expected.\n"), modifier->funName, 3);
        return sciErr.iErr;
    }

    sciErr = getVarAddressFromPosition(pvApiCtx, 4, &piAdressVr);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }

    vrHasAValidIntegerType = isIntegerType(pvApiCtx, piAdressVr);
    if (vrHasAValidIntegerType)
    {
        sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressVr, &vrIntegerType);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return sciErr.iErr;
        }
    }
    else
    {
        vrIntegerType = -1;
    }

    if (vrHasAValidIntegerType)
    {
        switch (vrIntegerType)
        {
            case SCI_INT32:
                sciErr = getMatrixOfInteger32(pvApiCtx, piAdressVr, &rows, &cols, (int**) vr);
                break;
            case SCI_UINT32:
                sciErr = getMatrixOfUnsignedInteger32(pvApiCtx, piAdressVr, &rows, &cols, vr);
                break;
            default:
                sciprint(_("%s: Wrong value for input argument #%d: An integer value expected.\n"), modifier->funName, 4);
                return 999;
                break;
        }
    }
    else
    {
        sciErr = getMatrixOfDoubleAsInteger(pvApiCtx, piAdressVr, &rows, &cols, (int**) vr);
    }
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }

    /* Store the rows and cols out */
    modifier->vrRows = rows;
    modifier->vrCols = cols;

    /* nvr is fixed by the vr length */
    *nvr = rows * cols;
    if (!modifier->calculateNvr)
    {
        /* use the nvr input arguments */
        size_t nvrArgument;

        sciErr = getVarAddressFromPosition(pvApiCtx, 5, &piAdressNvr);
        if (sciErr.iErr)
        {
            printError(&sciErr, 0);
            return sciErr.iErr;
        }

        nvrHasIntegerType = isIntegerType(pvApiCtx, piAdressNvr);
        if (nvrHasIntegerType)
        {
            sciErr = getMatrixOfIntegerPrecision(pvApiCtx, piAdressNvr, &nvrIntegerType);
            if (sciErr.iErr)
            {
                printError(&sciErr, 0);
                return sciErr.iErr;
            }
        }
        else
        {
            nvrIntegerType = -1;
        }

        if (nvrHasIntegerType)
        {
            int iNvrArgument;
            unsigned int uiNvrArgument;

            switch (nvrIntegerType)
            {
                case SCI_INT32:
                    sciErr.iErr = getScalarInteger32(pvApiCtx, piAdressNvr, &iNvrArgument);
                    nvrArgument = iNvrArgument;
                    break;
                case SCI_UINT32:
                    sciErr.iErr = getScalarUnsignedInteger32(pvApiCtx, piAdressNvr, &uiNvrArgument);
                    nvrArgument = uiNvrArgument;
                    break;
                default:
                    sciprint(_("%s: Wrong value for input argument #%d: An integer value expected.\n"), modifier->funName, 4);
                    return 999;
            }
        }
        else
        {
            double dNvrArgument;
            sciErr.iErr = getScalarDouble(pvApiCtx, piAdressNvr, &dNvrArgument);
            nvrArgument = (size_t) dNvrArgument;
        }
        if (sciErr.iErr)
        {
            return sciErr.iErr;
        }

        if (*nvr != nvrArgument)
        {
            sciprint(_("%s: Wrong value for input argument #%d: ""%d"" expected .\n"), "fmiCommonAccessors", 5, *nvr);
            return 999;
        }
    }

    return 0;
}

static int createEventInfoMList(void* _pvCtx, int _iVar, fmiEventInfo eventInfo)
{
    SciErr sciErr;

    int* piAddressEventInfo;
    int eventInfoScilabValue;

    static const char const* eventInfoHeader [] = {"fmiEventInfo", "iterationConverged", "stateReferencesChanged", "stateValuesChanged", "terminateSimulation", "upcomingTimeEvent", "nextEventTime"};

    sciErr = createMList(_pvCtx, _iVar, sizeof(eventInfoHeader) / sizeof(eventInfoHeader[0]), &piAddressEventInfo);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    sciErr = createMatrixOfStringInList(_pvCtx, _iVar, piAddressEventInfo, 1, 1, sizeof(eventInfoHeader) / sizeof(eventInfoHeader[0]), eventInfoHeader);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    eventInfoScilabValue = eventInfo.iterationConverged == fmiTrue ? TRUE : FALSE;
    sciErr = createMatrixOfBooleanInList(_pvCtx, _iVar, piAddressEventInfo, 2, 1, 1, &eventInfoScilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    eventInfoScilabValue = eventInfo.stateValueReferencesChanged == fmiTrue ? TRUE : FALSE;
    sciErr = createMatrixOfBooleanInList(_pvCtx, _iVar, piAddressEventInfo, 3, 1, 1, &eventInfoScilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    eventInfoScilabValue = eventInfo.stateValuesChanged == fmiTrue ? TRUE : FALSE;
    sciErr = createMatrixOfBooleanInList(_pvCtx, _iVar, piAddressEventInfo, 4, 1, 1, &eventInfoScilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    eventInfoScilabValue = eventInfo.terminateSimulation == fmiTrue ? TRUE : FALSE;
    sciErr = createMatrixOfBooleanInList(_pvCtx, _iVar, piAddressEventInfo, 5, 1, 1, &eventInfoScilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    eventInfoScilabValue = eventInfo.upcomingTimeEvent == fmiTrue ? TRUE : FALSE;
    sciErr = createMatrixOfBooleanInList(_pvCtx, _iVar, piAddressEventInfo, 6, 1, 1, &eventInfoScilabValue);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }
    sciErr = createMatrixOfDoubleInList(_pvCtx, _iVar, piAddressEventInfo, 7, 1, 1, &eventInfo.nextEventTime);
    if (sciErr.iErr)
    {
        printError(&sciErr, 0);
        return sciErr.iErr;
    }

    return 0;
}
