//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) - 2014 - Scilab Enterprises - Vladislav Trubkin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
//

#include "fmu_wrapper.h"
#include "api_scilab.h"
#include "Scierror.h"
#include "sciprint.h"
#include "scicos.h"
#include "scicos_block4.h"
#include "scicos_malloc.h"
#include "scicos_free.h"

void fmu_simulate_cs(scicos_block *block, scicos_flag flag)
{
    int pathSize = 0, identSize = 0, guidSize = 0, i = 0, j = 0; 
    char *pathName = NULL, *identName = NULL, *guid = NULL;
    double *sciPath = NULL, *sciIdent = NULL, *sciGuid = NULL; 
    double *sciInputTypes = NULL, *sciOutputTypes = NULL, *sciDebugLogging = NULL;
    fmiStatus status;
    // main structure
    workFMU *structFmu;
    fmiCallbackFunctions functions;
    double time = 0, dt = 0; 
    fmiValueReference *sciRealRefs = NULL, *sciIntRefs = NULL, *sciBoolRefs = NULL; 
    fmiValueReference *sciInputRefs = NULL, *sciOutputRefs = NULL; 
    fmiReal *rpar = NULL, *boolValue = NULL;
    double *valueRealSci = NULL, *valueBoolSci = NULL;
    int *valueIntSci = NULL;
    fmiInteger *ipar = NULL;
    size_t realRefsSize, intRefsSize, boolRefsSize, outRefsSize, inpRefsSize;
    fmiString fmuLocation = NULL, mimeType = NULL;
    fmiReal timeout = 1000, step = 0, tStart = 0, tStop = 0;
    fmiBoolean visible = fmiFalse, interactive = fmiFalse; 
    fmiBoolean loggingOn = fmiFalse, updateStep = fmiTrue;
    fmiBoolean StopTimeDefined = fmiFalse; 
    
    // return if an fmu has not been loaded
    if (block->nopar == 0) return;
    
    // real values
    sciRealRefs = Getuint32OparPtrs(block, 6); // real refs
    realRefsSize = GetOparSize(block, 6, 2); // real refs size
    rpar = GetRparPtrs(block); // real values 
    // integer values
    sciIntRefs = Getuint32OparPtrs(block, 7); 
    intRefsSize = GetOparSize(block, 7, 2); 
    ipar = GetIparPtrs(block); 
    // boolean values
    sciBoolRefs = Getuint32OparPtrs(block, 8); 
    boolRefsSize = GetOparSize(block, 8, 2);  
    boolValue = GetRealOparPtrs(block, 9); // boolean values
    // inputs 
    sciInputRefs = Getuint32OparPtrs(block, 2); 
    inpRefsSize = GetOparSize(block, 2, 2);
    sciInputTypes = GetRealOparPtrs(block, 3); 
    // outputs
    sciOutputRefs = Getuint32OparPtrs(block, 4);
    outRefsSize = GetOparSize(block, 4, 2);
    sciOutputTypes = GetRealOparPtrs(block, 5);
    
    if (flag == Initialization)
    {
        // sciprint("\n");
        // sciprint("*****Initialization of block [%d] Time [%f] \n",get_block_number(), get_scicos_time());
        pathSize = GetOparSize(block, 10, 1); // model.opar(10); path for library
        sciPath = GetRealOparPtrs(block, 10); // decimal values
        identSize = GetOparSize(block, 11, 1); 
        sciIdent = GetRealOparPtrs(block, 11); // identifier
        guidSize = GetOparSize(block, 12, 1);
        sciGuid = GetRealOparPtrs(block, 12); // guid
        sciDebugLogging = GetRealOparPtrs(block, 13); // logger
        loggingOn = (fmiBoolean)sciDebugLogging[0];

        // memory allocation for the main structure 
        if ((*(block->work) = (workFMU*)scicos_malloc(sizeof(workFMU))) == NULL)
        {
            set_block_error(-16); // memory allocation error
            *(block->work) = NULL; 
            return;
        }
        structFmu = *(block->work); 
        // path
        if ((pathName = (char*)scicos_malloc(pathSize * sizeof(char))) == NULL)
        {
            set_block_error(-16);
            scicos_free(structFmu);
            *(block->work) = NULL;
            return;
        }
        else 
        {
        memset(pathName, 0x0, pathSize * sizeof(char));
        }
        for (i=0; i<pathSize; i++)
        {
        pathName[i] = (char)sciPath[i];
        }
        // identifier
        if ((identName = (char*)scicos_malloc(identSize * sizeof(char))) == NULL)
        {
            set_block_error(-16);
            scicos_free(structFmu);
            *(block->work) = NULL;
            return;
        }
        else
        {
            memset(identName, 0x0, identSize * sizeof(char));
        }
        for (i=0; i<identSize; i++)
        {
            identName[i] = (char)sciIdent[i];
        } 
        // guid
        if ((guid = (char*)scicos_malloc(guidSize * sizeof(char))) == NULL)
        {
            set_block_error(-16); // memory allocation 
            scicos_free(structFmu);
            *(block->work) = NULL;
            return;
        }
        else
        {
            memset(guid, 0x0, guidSize * sizeof(char));
        }			
        for (i=0; i<guidSize; i++)
        {
            guid[i] = (char)sciGuid[i];
        } 
        
        // load library and create an instance 
        if ((structFmu->fmu = (void*)fmu_wrap_co_simulation_new(pathName, identName)) == NULL)
        {
            sciprint(fmu_wrap_error(), "fmu_simulate_cs");
            free_instance(block, -3, "fmu_wrap_co_simulation_new"); // internal error
            return;
        }
        // set fmiType
        structFmu->type = CoSimulation;
        
        // helper functions
        functions.logger = fmu_callback_logger;
        functions.allocateMemory = calloc;
        functions.freeMemory = free;
        functions.stepFinished = NULL;
        
        // instantiate slave
        structFmu->instance = ((FmuCoSimulation*)structFmu->fmu)->fmiInstantiateSlave(identName, guid, fmuLocation, mimeType, timeout, visible, interactive, functions, loggingOn);
        if (structFmu->instance == NULL)
        {
            free_instance(block, -3, "fmiInstantiateSlave");
            return;
        }
        // set debug logging
        if (loggingOn)
        {
           status = ((FmuCoSimulation*)structFmu->fmu)->fmiSetDebugLogging(structFmu->instance, loggingOn);
            if (status > fmiWarning)
            {
                free_instance(block, -3, "fmiSetDebugLogging");
                return;
            } 
        }
        
        // set start values
        // real
        if (realRefsSize > 0)
        {
            if (((structFmu->realRefs) = (fmiValueReference*)scicos_malloc(realRefsSize * sizeof(fmiValueReference))) == NULL)
            {
                free_instance(block, -16, "scicos_malloc");
                return;
            }
            // negated alias
            for (i=0, j=1; i<realRefsSize; i++)
            {
                if (sciRealRefs[j] > 0) 
                {
                    rpar[i] = -rpar[i];
                }
                j += 2;	
            }	
            for (j=0, i=0; i<realRefsSize; i++)
            {
                sciRealRefs[i] = sciRealRefs[j];
                structFmu->realRefs[i] = sciRealRefs[i]; 
                j += 2;
            }
            // set real values	
            status = ((FmuCoSimulation*)structFmu->fmu)->fmiSetReal((structFmu->instance), structFmu->realRefs, realRefsSize, rpar); 
            if (status > fmiDiscard) 
            {
                free_instance(block, -3, "fmiSetReal");
                return;
            }
        }	
        // integers
        if (intRefsSize > 0)
        {
            if (((structFmu->intRefs) = (fmiValueReference*)scicos_malloc(intRefsSize * sizeof(fmiValueReference))) == NULL)
            {
                free_instance(block, -16, "scicos_malloc");
                return;
            }	
            // negated alias
            for (i=0, j=1; i<intRefsSize; i++)
            {
                if (sciIntRefs[j] > 0) 
                {
                    ipar[i] = -ipar[i];
                }
                j += 2;	
            }
            for (j=0, i=0; i<intRefsSize; i++)
            {
                sciIntRefs[i] = sciIntRefs[j];
                structFmu->intRefs[i] = sciIntRefs[i]; 
                j += 2;
            }
            // set integer values	
            status = ((FmuCoSimulation*)structFmu->fmu)->fmiSetInteger((structFmu->instance), structFmu->intRefs, intRefsSize, ipar); 
            if (status > fmiWarning) 
            {
                free_instance(block, -3, "fmiSetInteger");
            }
        }
        // boolean
        if (boolRefsSize > 0)
        {
            if (((structFmu->boolValue) = (fmiBoolean*)scicos_malloc(boolRefsSize * sizeof(fmiBoolean))) == NULL)
            {
                free_instance(block, -16, "scicos_malloc");
                return;
            }	
            for (i=0, j=1; i<boolRefsSize; i++)
            {
                if (sciBoolRefs[j] > 0) 
                {
                    // negated alias
                    if (boolValue[i] == 1) 
                    {
                        structFmu->boolValue[i] = fmiFalse;
                    }
                    else
                    {
                        structFmu->boolValue[i] = fmiTrue;
                    }
                }
                else 
                {
                    if (boolValue[i] == 1) 
                    {
                        structFmu->boolValue[i] = fmiTrue;
                    }
                    else
                    {
                        structFmu->boolValue[i] = fmiFalse;
                    }
                }
                j += 2;	
            }
            if (((structFmu->boolRefs) = (fmiValueReference*)scicos_malloc(boolRefsSize * sizeof(fmiValueReference))) == NULL)
            {
                free_instance(block, -16, "scicos_malloc");
                return;
            }	
            for (j=0, i=0; i<boolRefsSize; i++)
            {
                sciBoolRefs[i] = sciBoolRefs[j];
                structFmu->boolRefs[i] = sciBoolRefs[i]; 
                j += 2;
            }
            // set boolean values	
            status = ((FmuCoSimulation*)structFmu->fmu)->fmiSetBoolean((structFmu->instance), structFmu->boolRefs, boolRefsSize, structFmu->boolValue);
            if (status > fmiWarning) 
            {
                free_instance(block, -3, "fmiSetBoolean");
                return;
            }
        }
        
        // initialize slave
        tStart = get_scicos_time();
        // StopTimeDefined == fmiFalse, tStop == 0
        status = ((FmuCoSimulation*)structFmu->fmu)->fmiInitializeSlave((structFmu->instance), tStart, StopTimeDefined, tStop);
        if (status > 1)
        {
            free_instance(block, -3, "fmiInitializeSlave");
            return;
        }
        set_values(block, sciInputTypes, inpRefsSize, sciInputRefs);
    }
    
    else if (flag == OutputUpdate)
	{
        // sciprint("\n");
        // sciprint("*****Output update block [%d] Time [%f] \n",get_block_number(), get_scicos_time());
        if ((structFmu = *(block->work)) == NULL) return;
        time = get_scicos_time();
        
        dt = get_scicos_time() - structFmu->oldTime;
        // save previous time
        structFmu->oldTime = get_scicos_time();
        
        if ((dt == 0 || dt < 0) && time != 0) 
        {
            updateStep = fmiFalse;
        }
        else if (time == 0)
        {
            step = get_scicos_time();
        }
        else
        {
            step = dt;
        }
        
        if (updateStep) 
        { 
            // do step
            status = ((FmuCoSimulation*)structFmu->fmu)->fmiDoStep((structFmu->instance), time-step, step, fmiTrue);
            if (status > fmiDiscard)
            {
                free_instance(block, -3, "fmiDoStep");
                return;
            }
            // get outputs/set inputs
            get_values(block, sciOutputTypes, outRefsSize, sciOutputRefs);
            set_values(block, sciInputTypes, inpRefsSize, sciInputRefs);
        }
	}
    
    else if (flag == Ending)
	{
		// sciprint("\n");
		// sciprint("*****Ending of block [%d] Time [%f] \n", get_block_number(), get_scicos_time());
		if ((structFmu = *(block->work)) == NULL) return;
			
        // terminate simulation
        status = ((FmuCoSimulation*)structFmu->fmu)->fmiTerminateSlave(structFmu->instance);
        if (status > fmiWarning)
        {
            free_instance(block, -3, "fmiTerminateSlave");
            return;
        }
        
        // freeing of instance
        ((FmuCoSimulation*)structFmu->fmu)->fmiFreeSlaveInstance(structFmu->instance);
        // unload library and delete all
        fmu_wrap_co_simulation_delete((FmuCoSimulation*)structFmu->fmu);
        scicos_free(structFmu);
	}
}