//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) - 2013 - Scilab Enterprises - Clement David
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
//

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#ifdef _MSC_VER
#include <windows.h>
#else
#include <dlfcn.h>
#endif

#ifdef _MSC_VER
#define SNPRINTF sprintf_s
#else
#define SNPRINTF snprintf
#endif

#include "fmu_wrapper.h"

#define MAX_FMU_NAME_LEN 128
#define MAX_ERR_LEN 256

/*
 * Helper functions
 */
static void* fmu_wrap_dlopen(const char* lib);
static void* fmu_wrap_dlsym(void* sharedLibraryHandle, const char* symbol);
static int fmu_wrap_dlclose(void* sharedLibraryHandle);
static void* fmu_wrap_map_symbol(void* sharedLibraryHandle, char* symbolBufferedName, size_t prefixLen, const char* name);

static const char const* MODEL_EXCHANGE_KLASS = "FmuModelExchange";
static const char const* CO_SIMULATION_KLASS = "FmuCoSimulation";
static char fmu_wrap_err[MAX_ERR_LEN];

FmuModelExchange* fmu_wrap_model_exchange_new(const char* lib, const char* name)
{
    void* sharedLibraryHandle;
    char* symbolBufferedName;

    FmuModelExchange* instance;
    size_t prefixLen;

    /*
     * load the dynamic library and function
     */
    sharedLibraryHandle = fmu_wrap_dlopen(lib);
    if (sharedLibraryHandle == NULL)
    {
        return NULL;
    }

    /*
     * Allocate and init the instance
     */
    instance = calloc(1, sizeof(FmuModelExchange));
    if (instance == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_sharedLibraryHandle;
    }
    instance->klass = MODEL_EXCHANGE_KLASS;

    prefixLen = strlen(name);
    if (prefixLen > (MAX_FMU_NAME_LEN / 2))
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: FMI model name is too long.\n"), "%s");
        goto err_free_instance;
    }

    instance->klassName = malloc(prefixLen + 1);
    if (instance == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_instance;
    }
    memcpy(instance->klassName, name, prefixLen + 1);
    instance->sharedLibraryHandle = sharedLibraryHandle;

    /*
     * allocate the re-useable symbol fullname
     */
    symbolBufferedName = malloc((prefixLen + 1 + MAX_FMU_NAME_LEN + 1) * sizeof(char));
    if (symbolBufferedName == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_klassName;
    }
    memcpy(symbolBufferedName, name, prefixLen);
    symbolBufferedName[prefixLen] = '_';
    symbolBufferedName[prefixLen + 1] = '\0';

    /*
     * Clear the error variable then map all the symbols using macros
     *
     * Error tracking is performed at the end of the mapping accordingly to the error variable
     *
     */
    fmu_wrap_err[0] = '\0';

#define FMI_KLASS_METHOD(fun) instance->fun = fmu_wrap_map_symbol(sharedLibraryHandle, symbolBufferedName, prefixLen, #fun);
    FMI_KLASS_METHOD(fmiGetModelTypesPlatform)
    FMI_KLASS_METHOD(fmiGetVersion)
    FMI_KLASS_METHOD(fmiInstantiateModel)
    FMI_KLASS_METHOD(fmiFreeModelInstance)
    FMI_KLASS_METHOD(fmiSetDebugLogging)
    FMI_KLASS_METHOD(fmiSetTime)
    FMI_KLASS_METHOD(fmiSetContinuousStates)
    FMI_KLASS_METHOD(fmiCompletedIntegratorStep)
    FMI_KLASS_METHOD(fmiSetReal)
    FMI_KLASS_METHOD(fmiSetInteger)
    FMI_KLASS_METHOD(fmiSetBoolean)
    FMI_KLASS_METHOD(fmiSetString)
    FMI_KLASS_METHOD(fmiInitialize)
    FMI_KLASS_METHOD(fmiGetDerivatives)
    FMI_KLASS_METHOD(fmiGetEventIndicators)
    FMI_KLASS_METHOD(fmiGetReal)
    FMI_KLASS_METHOD(fmiGetInteger)
    FMI_KLASS_METHOD(fmiGetBoolean)
    FMI_KLASS_METHOD(fmiGetString)
    FMI_KLASS_METHOD(fmiEventUpdate)
    FMI_KLASS_METHOD(fmiGetContinuousStates)
    FMI_KLASS_METHOD(fmiGetNominalContinuousStates)
    FMI_KLASS_METHOD(fmiGetStateValueReferences)
    FMI_KLASS_METHOD(fmiTerminate)
#undef FMI_KLASS_METHOD

    free(symbolBufferedName);

    /*
     * On a successful mapping, return the instance
     */
    if (fmu_wrap_err[0] == '\0')
    {
		return instance;
    }
err_free_klassName:
    free(instance->klassName);
err_free_instance:
    free(instance);
err_free_sharedLibraryHandle:
    fmu_wrap_dlclose(sharedLibraryHandle);
    return NULL;
}

FmuCoSimulation* fmu_wrap_co_simulation_new(const char* lib, const char* name)
{
    void* sharedLibraryHandle;
    char* symbolBufferedName;

    FmuCoSimulation* instance;
    size_t prefixLen;

    /*
     * load the dynamic library and function
     */
    sharedLibraryHandle = fmu_wrap_dlopen(lib);
    if (sharedLibraryHandle == NULL)
    {
        return NULL;
    }

    /*
     * Allocate and init the instance
     */
    instance = calloc(1, sizeof(FmuCoSimulation));
    if (instance == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_sharedLibraryHandle;
    }
    instance->klass = CO_SIMULATION_KLASS;

    prefixLen = strlen(name);
    if (prefixLen > (MAX_FMU_NAME_LEN / 2))
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: FMI model name is too long.\n"), "%s");
        goto err_free_instance;
    }

    instance->klassName = malloc(prefixLen + 1);
    if (instance == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_instance;
    }
    memcpy(instance->klassName, name, prefixLen + 1);
    instance->sharedLibraryHandle = sharedLibraryHandle;

    /*
     * allocate the re-useable symbol fullname
     */
    symbolBufferedName = malloc((prefixLen + 1 + MAX_FMU_NAME_LEN + 1) * sizeof(char));
    if (symbolBufferedName == NULL)
    {
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: Cannot allocate memory.\n"), "%s");
        goto err_free_klassName;
    }
    memcpy(symbolBufferedName, name, prefixLen);
    symbolBufferedName[prefixLen] = '_';
    symbolBufferedName[prefixLen + 1] = '\0';

    /*
     * Clear the error variable then map all the symbols using macros
     *
     * Error tracking is performed at the end of the mapping accordingly to the error variable
     *
     */
    fmu_wrap_err[0] = '\0';

#define FMI_KLASS_METHOD(fun) instance->fun = fmu_wrap_map_symbol(sharedLibraryHandle, symbolBufferedName, prefixLen, #fun);
    FMI_KLASS_METHOD(fmiGetTypesPlatform)
    FMI_KLASS_METHOD(fmiGetVersion)
    FMI_KLASS_METHOD(fmiSetDebugLogging)
    FMI_KLASS_METHOD(fmiSetReal)
    FMI_KLASS_METHOD(fmiSetInteger)
    FMI_KLASS_METHOD(fmiSetBoolean)
    FMI_KLASS_METHOD(fmiSetString)
    FMI_KLASS_METHOD(fmiGetReal)
    FMI_KLASS_METHOD(fmiGetInteger)
    FMI_KLASS_METHOD(fmiGetBoolean)
    FMI_KLASS_METHOD(fmiGetString)
    FMI_KLASS_METHOD(fmiInstantiateSlave)
    FMI_KLASS_METHOD(fmiInitializeSlave)
    FMI_KLASS_METHOD(fmiTerminateSlave)
    FMI_KLASS_METHOD(fmiResetSlave)
    FMI_KLASS_METHOD(fmiFreeSlaveInstance)
    FMI_KLASS_METHOD(fmiSetRealInputDerivatives)
    FMI_KLASS_METHOD(fmiGetRealOutputDerivatives)
    FMI_KLASS_METHOD(fmiDoStep)
    FMI_KLASS_METHOD(fmiCancelStep)
    FMI_KLASS_METHOD(fmiGetStatus)
    FMI_KLASS_METHOD(fmiGetRealStatus)
    FMI_KLASS_METHOD(fmiGetIntegerStatus)
    FMI_KLASS_METHOD(fmiGetBooleanStatus)
    FMI_KLASS_METHOD(fmiGetStringStatus)
#undef FMI_KLASS_METHOD

    free(symbolBufferedName);

    /*
     * On a successful mapping, return the instance
     */
    if (fmu_wrap_err[0] == '\0')
    {
		return instance;
    }
err_free_klassName:
    free(instance->klassName);
err_free_instance:
    free(instance);
err_free_sharedLibraryHandle:
    fmu_wrap_dlclose(sharedLibraryHandle);
    return NULL;
}

void fmu_wrap_model_exchange_delete(FmuModelExchange* instance)
{
    void* sharedLibraryHandle;

    sharedLibraryHandle = instance->sharedLibraryHandle;
    free(instance->klassName);
    free(instance);
    fmu_wrap_dlclose(sharedLibraryHandle);
}

void fmu_wrap_co_simulation_delete(FmuCoSimulation* instance)
{
    void* sharedLibraryHandle;

    sharedLibraryHandle = instance->sharedLibraryHandle;
    free(instance->klassName);
    free(instance);
    fmu_wrap_dlclose(sharedLibraryHandle);
}

const char* fmu_wrap_error()
{
    return fmu_wrap_err;
}

FmuModelExchange* fmu_wrap_model_exchange(void* instance)
{
    FmuModelExchange* local = (FmuModelExchange*) instance;

    if (local->klass != MODEL_EXCHANGE_KLASS)
    {
        return NULL;
    }

    return local;
}

FmuCoSimulation* fmu_wrap_co_simulation(void* instance)
{
    FmuCoSimulation* local = (FmuCoSimulation*) instance;

    if (local->klass != CO_SIMULATION_KLASS)
    {
        return NULL;
    }

    return local;
}

/*
 * Helper functions implementation
 */

static void* fmu_wrap_dlopen(const char* lib)
{
    void *sharedLibraryHandle;
    char *error;

#ifdef _MSC_VER
    sharedLibraryHandle = LoadLibrary (lib);
#else
    sharedLibraryHandle = dlopen (lib, RTLD_LAZY | RTLD_LOCAL);
#endif

    if (sharedLibraryHandle == NULL)
    {
#ifdef _MSC_VER
        error = GetLastDynLibError ();
#else
        error = dlerror ();
#endif
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: The shared archive was not loaded: %s\n"), "%s" , error);
    }

    return sharedLibraryHandle;
}

static void* fmu_wrap_dlsym(void* sharedLibraryHandle, const char* symbol)
{
    void* function_ptr;
    char *error;

#ifdef _MSC_VER
    function_ptr = GetProcAddress (sharedLibraryHandle, symbol);
#else
    dlerror (); /* clear the last error */
    function_ptr = dlsym (sharedLibraryHandle, symbol);
#endif

    if (function_ptr == NULL)
    {
#ifdef _MSC_VER
        error = GetLastDynLibError ();
#else
        error = dlerror ();
#endif
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: %s is not an entry point: %s\n"), "%s", symbol, error);
    }

    return function_ptr;
}

static int fmu_wrap_dlclose(void* sharedLibraryHandle)
{
    char *error;
    int err;

#ifdef _MSC_VER
    err = FreeLibrary (sharedLibraryHandle);
#else
    err = dlclose (sharedLibraryHandle);
#endif

    if (err)
    {
#ifdef _MSC_VER
        error = GetLastDynLibError ();
#else
        error = dlerror ();
#endif
        SNPRINTF(fmu_wrap_err, MAX_ERR_LEN, _("%s: The shared archive was not unloaded: %s\n"), "%s", error);
    }

    return err;
}

static void* fmu_wrap_map_symbol(void* sharedLibraryHandle, char* symbolBufferedName, size_t prefixLen, const char* name)
{
    /* to ease error handling do not try to map when an error is present */
    if (fmu_wrap_err[0] != '\0')
    {
        return NULL;
    }

    /* use the buffer to construct the full name*/
    memcpy(symbolBufferedName + prefixLen + 1, name, MAX_FMU_NAME_LEN);
    symbolBufferedName[prefixLen + 1 + MAX_FMU_NAME_LEN] = '\0';

    /* use the cross-platform dlsym to resolve the symbol */
    return fmu_wrap_dlsym(sharedLibraryHandle, symbolBufferedName);
}

/**
 * log some information into the Scilab console
*/
void fmu_callback_logger(fmiComponent c, fmiString instanceName, fmiStatus status, fmiString category, fmiString message, ...)
{
    va_list ap;

    if (!instanceName)
    {
        instanceName = "?";
    }
    if (!category)
    {
        category = "?";
    }
    
    sciprint("fmiStatus = %d\n", status);
    sciprint("%s (%s)\n", instanceName, category);
    va_start(ap, message);


    scivprint(message, ap);
    va_end(ap);
    
	/* FMI for ModelExchange v1.0: "logger function will append a line break to each message" */

    sciprint("\n");
}
